# coding=utf-8
"""Implementation of a Directed Graph data structure."""
from __future__ import division
import math

from ladybug_geometry.geometry2d import LineSegment2D, Polygon2D
from ladybug_geometry.intersection2d import intersect_line2d_infinite

from .polyskel import skeleton_as_edge_list, _intersect_skeleton_segments, \
    _remove_segments_outside_boundary


def _vector2hash(vector, tol):
    """Hashes spatial coordinates for use in dictionary.

    Args:
        vector: A Vector2D object.
        tol: floating point precision tolerance.

    Returns:
        Hash of vector as a string of rounded coordinates.
    """
    # get the relative tolerance using a log function
    try:
        rtol = int(math.log10(tol)) * -1
    except ValueError:
        rtol = 0  # the tol is equal to 1 (out of range for log)
    # account for the fact that the tolerance may not be base 10
    base = int(tol * 10 ** (rtol + 1))
    if base == 10 or base == 0:  # tolerance is base 10 (eg. 0.001)
        base = 1
    else:  # tolerance is not base 10 (eg. 0.003)
        rtol += 1
    # avoid cases of signed zeros messing with the hash
    x_val = 0.0 if vector.x == 0 else vector.x
    y_val = 0.0 if vector.y == 0 else vector.y
    # convert the coordinate values to a hash
    return str((
        base * round(x_val / base, rtol),
        base * round(y_val / base, rtol)
    ))


class _Node(object):
    """Private class to handle nodes in PolygonDirectedGraph.

    Args:
        val: A Point2D object.
        key: Hash of Point2D object.
        order: integer counting order of the Node (based on dg propagation)
        adj_lst: list of keys adjacent to this node.
        exterior: Node boundary condition. None if not set by user, else True
            or False according to user.

    Properties:
        * pt
        * key
        * adj_lst
        * exterior
        * adj_count
    """
    __slots__ = ('key', 'pt', '_order', 'adj_lst', 'exterior')

    def __init__(self, key, val, order, adj_lst, exterior):
        """Initialize _Node"""

        self.key = key
        self.pt = val
        self._order = order
        self.adj_lst = adj_lst
        # Potentially change exterior to data (similar to networkX)
        # and pass conditional function to get_exterior
        # this resolves redundancy between unidirect and exterior
        # node/edge properties.
        self.exterior = exterior

    @property
    def adj_count(self):
        """Number of adjacent nodes"""
        return len(self.adj_lst)

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return isinstance(other, _Node) and \
            self.key == other.key

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{}: {}'.format(self._order, self.key)


class PolygonDirectedGraph(object):
    """A directed graph data structure for point relationships.

    The PolygonDirectedGraph effectively represents the network of a straight
    skeleton and assists with finding the shortest pathways through it. It also
    helps differentiate interior from exterior parts of the graph. Typically,
    interior pathways are bi-directional in the graph while exterior pathways
    are uni-directional.

    Args:
        tolerance: Tolerance for point equivalence. (Default: 1e-5). This is used
            for hashing points within the network.

    Properties:
        * node_count: Integer for the number of nodes in graph.
        * nodes: An iterable of nodes in graph.
        * ordered_nodes: An iterable of nodes in graph in order they were added.
        * connection_segments: List of LineSegment2D for the node connections
        * outer_root_node: A node for the outer root key
        * hole_root_nodes: A list of nodes for the hole root keys
        * is_intersect_topology: A boolean for whether the graph self-intersects
    """

    def __init__(self, tol):
        """Initialize a PolygonDirectedGraph."""
        self._directed_graph = {}
        self._tol = tol
        self.outer_root_key = None  # will be set during skeleton creation
        self.hole_root_keys = []  # will be set during skeleton creation
        self.is_intersect_topology = None  # will be set during skeleton creation

    @classmethod
    def from_polygon(cls, polygon, tol):
        """Generate a directed graph from a polygon.

        Args:
            polygon: A Polygon2D object.
        """
        return cls.from_point_array(polygon.vertices, tol, loop=True)

    @classmethod
    def from_point_array(cls, point_array, tol, loop=True):
        """Generate a directed graph from a 1-dimensional array of points.

        Args:
            point_array: Array of Point2D objects.
            loop: Optional parameter to connect 1d array
            tol: Tolerance for point equivalence.
        """
        dg = cls(tol)
        for i in range(len(point_array) - 1):
            dg.add_node(point_array[i], [point_array[i + 1]], exterior=None)
        if loop:
            dg.add_node(point_array[-1], [point_array[0]], exterior=None)
        return dg

    @property
    def node_count(self):
        return len(self.nodes)

    @property
    def nodes(self):
        """Get an iterable of pt nodes"""
        return self._directed_graph.values()

    @property
    def ordered_nodes(self):
        """Get an iterable of pt nodes in order of addition"""
        nodes = list(self.nodes)
        nodes.sort(key=lambda v: v._order)
        return nodes

    @property
    def outer_root_node(self):
        """Get the node of the outer boundary root."""
        return self.node(self.outer_root_key)

    @property
    def hole_root_nodes(self):
        """Get a list of nodes for the roots of the holes."""
        return [self.node(hole_key) for hole_key in self.hole_root_keys]

    @property
    def connection_segments(self):
        """Get a list of LineSegment2D for the node connections in the graph."""
        traversed = set()
        connections = []
        for node in self.nodes:
            for conn_node in node.adj_lst:
                if (conn_node.key, node.key) not in traversed:
                    conn_seg = LineSegment2D.from_end_points(node.pt, conn_node.pt)
                    connections.append(conn_seg)
                    traversed.add((node.key, conn_node.key))
        return connections

    def node(self, key):
        """Retrieves the node based on passed value.

        Args:
            val: The key for a node in the directed graph.

        Returns:
            The node for the passed key.
        """
        try:
            return self._directed_graph[key]
        except KeyError:
            return None  # broken connection

    def add_adj(self, node, adj_val_lst):
        """Adds nodes to node.adj_lst.

        This method will ensure no repetitions will occur in adj_lst.

        Args:
            node: _Node to add adjacencies to.
            adj_val_lst: List of Point2D objects to add as adjacent nodes.
        """
        adj_keys = {n.key: None for n in node.adj_lst}
        adj_keys[node.key] = None
        for adj_val in adj_val_lst:
            adj_key = _vector2hash(adj_val, self._tol)
            if adj_key in adj_keys:
                continue

            self._add_node(adj_key, adj_val, exterior=None)
            adj_keys[adj_key] = None
            node.adj_lst.append(self.node(adj_key))

    def remove_adj(self, node, adj_key_lst):
        """Removes nodes in node.adj_lst.

        Args:
            node: _Node to remove adjacencies to.
            adj_val_lst: List of adjacency keys to remove as adjacent nodes.
        """
        node.adj_lst = [n for n in node.adj_lst if n.key not in set(adj_key_lst)]

    def add_node(self, val, adj_lst, exterior=None):
        """Add a node into the PolygonDirectedGraph.

        This method consumes a Point2D, computes its key value, and adds it in the
        graph if it doesn't exist. If it does exist it appends adj_lst to existing pt.

        Args:
            val: A Point2D object.
            adj_lst: A list of Point2D objects adjacent to the node.
            exterior: Optional boolean for exterior attribute.

        Returns:
            The hashed key from the existing or new node.
        """
        key = _vector2hash(val, self._tol)  # get key
        self._add_node(key, val, exterior)  # get node if it exists
        node = self._directed_graph[key]
        self.add_adj(node, adj_lst)  # add the adj_lst to dg
        # if the exterior boolean was passed, change the node attribute
        if exterior is not None:
            node.exterior = exterior
        return node.key

    def _add_node(self, key, val, exterior=None):
        """Helper function for add_node. If key doesn't exist, add to dg."""
        if key not in self._directed_graph:
            self._directed_graph[key] = _Node(key, val, self.node_count, [], exterior)
        return self._directed_graph[key]

    def insert_node(self, base_node, new_val, next_node, exterior=None):
        """Insert node in the middle of an edge defined by node and next_node.

        Args:
            base_node: _Node object to the left.
            new_val:  A Point2D object for the new node in the middle.
            next_node: _Node object to the right.
            exterior: Optional boolean for exterior attribute.

        Returns:
            key of new_val node.
        """
        # add new_val as a node, with next_node as an adjacency
        new_key = self.add_node(new_val, [next_node.pt], exterior=exterior)
        # update parent by adding new adjacency, and removing old adjacency
        self.add_adj(base_node, [self.node(new_key).pt])

        # catch the edge case where the new point is coincident to parent or next_point.
        # this occurs when intersection passes through a corner.
        if (new_key == next_node.key) or (new_key == base_node.key):
            return new_key
        self.remove_adj(base_node, [next_node.key])
        return new_key

    def node_exists(self, key):
        """Check if a node is in the graph. True if node in directed graph else False."""
        return key in self._directed_graph

    def pt_exists(self, pt):
        """True if a point (as Point2D) in directed graph exists as node else False.
        """
        return self.node_exists(_vector2hash(pt, self._tol))

    def polygon_exists(self, polygon):
        """Check if polygon is in the directed graph.

        Args:
            polygons: A Polygon2D object.
            dg: A PolygonDirectedGraph.

        Return:
            True if exists, else False.
        """
        vertices_loop = list(polygon.vertices)
        vertices_loop = vertices_loop + [vertices_loop[0]]

        for i in range(len(vertices_loop) - 1):
            pt1 = vertices_loop[i]
            pt2 = vertices_loop[i + 1]

            if not self.pt_exists(pt1):
                return False

            node1 = self.node(_vector2hash(pt1, self._tol))
            node2 = self.node(_vector2hash(pt2, self._tol))
            if node2.key in [n.key for n in node1.adj_lst]:
                return False

        return True

    def adj_matrix(self):
        """Gets an adjacency matrix of the directed graph where:

        * 1 = adjacency from row node to col node.
        * 0 = no adjacency.

        Returns:
            N x N square matrix where N is number of nodes.
        """
        nodes = self.ordered_nodes
        # initialize a mtx with no adjacencies
        amtx = [[0 for i in range(self.node_count)]
                for j in range(self.node_count)]

        for i in range(self.node_count):
            adj_indices = [adj._order for adj in nodes[i].adj_lst]
            for adj_idx in adj_indices:
                amtx[i][adj_idx] = 1

        return amtx

    def adj_matrix_labels(self):
        """Returns a dictionary where label key corresponds to index in adj_matrix
        and value is node key"""
        return {i: node.key for i, node in enumerate(self.ordered_nodes)}

    def intersect_graph_with_segment(self, segment):
        """Update graph with intersection of partial segment crossing through polygon.

        Args:
            segment: LineSegment2D to intersect with the graph. The LineSegment2D
                does not need to be contained within polygon and will be intersected
                infinitely.
        """
        # loop through all nodes in the graph and find intersection points
        int_key_lst = []
        for node in self.ordered_nodes:
            # convert graph edge to trimming segment
            next_node = node.adj_lst[0]
            trim_seg = LineSegment2D.from_end_points(node.pt, next_node.pt)
            int_pt = intersect_line2d_infinite(trim_seg, segment)

            # add intersection point as new node in graph
            if int_pt:
                int_key = self.insert_node(node, int_pt, next_node, exterior=False)
                int_key_lst.append(int_key)

        # add intersection edges between the newly-found nodes
        if len(int_key_lst) == 2:
            # typical case with convex cases
            # make edge between intersection nodes
            n1, n2 = self.node(int_key_lst[0]), self.node(int_key_lst[1])
            self.add_node(n1.pt, [n2.pt], exterior=False)
            self.add_node(n2.pt, [n1.pt], exterior=False)

        elif len(int_key_lst) > 2:
            # edge case with concave geometry creates multiple intersections
            # sort distance and add adjacency
            n = self.node(int_key_lst[0])
            distances = [(0, 0.0)]

            for i, k in enumerate(int_key_lst[1:]):
                distance = LineSegment2D.from_end_points(n.pt, self.node(k).pt).length
                distances.append((i + 1, distance))

            distances = sorted(distances, key=lambda t: t[1])

            for i in range(len(distances) - 1):
                k1, k2 = distances[i][0], distances[i + 1][0]
                n1, n2 = self.node(int_key_lst[k1]), self.node(int_key_lst[k2])
                # add bi-direction so the min cycle works
                self.add_node(n1.pt, [n2.pt], exterior=False)
                self.add_node(n2.pt, [n1.pt], exterior=False)

    def min_cycle(self, base_node, goal_node, ccw_only=False):
        """Identify the shortest interior cycle between two exterior nodes.

        Args:
            base_node: The first exterior node of the edge.
            goal_node: The end exterior node of the cycle that, together with
                the base_node, constitutes an exterior edge.
            ccw_only: A boolean to note whether the search should be limited
                to the counter-clockwise direction only. (Default: False).

        Returns:
            A list of nodes that form a polygon if the cycle exists, else None.
        """
        # set up a queue for exploring the graph
        explored = []
        queue = [[base_node]]
        orig_dir = base_node.pt - goal_node.pt  # yields a vector
        # loop to traverse the graph  with the help of the queue
        while queue:
            path = queue.pop(0)
            node = path[-1]
            # make sure that the current node has not been visited
            if node not in explored:
                prev_dir = node.pt - path[-2].pt if len(path) > 1 else orig_dir
                # iterate over the neighbors to determine relevant nodes
                rel_neighbors, rel_angles = [], []
                for neighbor in node.adj_lst:
                    if neighbor == goal_node:  # the shortest path was found!
                        path.append(goal_node)
                        return path
                    elif neighbor.exterior:
                        continue  # don't traverse the graph exterior
                    edge_dir = neighbor.pt - node.pt
                    cw_angle = prev_dir.angle_clockwise(edge_dir * -1)
                    if not (1e-5 < cw_angle < (2 * math.pi) - 1e-5):
                        continue  # prevent back-tracking along the search
                    rel_neighbors.append(neighbor)
                    rel_angles.append(cw_angle)
                # sort the neighbors by clockwise angle
                if len(rel_neighbors) > 1:
                    rel_neighbors = [n for _, n in sorted(zip(rel_angles, rel_neighbors),
                                                          key=lambda pair: pair[0])]
                # add the relevant neighbors to the path and the queue
                if ccw_only:
                    new_path = list(path)
                    new_path.append(rel_neighbors[0])
                    queue.append(new_path)
                else:  # add all neighbors to the search
                    for neighbor in rel_neighbors:
                        new_path = list(path)
                        new_path.append(neighbor)
                        queue.append(new_path)
                explored.append(node)
        # if we reached the end of the queue, then no path was found
        return None

    def exterior_cycle(self, cycle_root):
        """Computes exterior boundary from a given node.

        This method assumes that exterior edges are naked (unidirectional) and
        interior edges are bidirectional.

        Args:
            cycle_root: Starting _Node in exterior cycle.

        Returns:
            List of nodes on exterior if a cycle exists, else None.
        """
        # Get the first exterior edge
        curr_node = cycle_root
        next_node = PolygonDirectedGraph.next_exterior_node(curr_node)
        if not next_node:
            return None

        # loop through the cycle until we get it all or run out of points
        max_iter = self.node_count + 1  # maximum length a cycle can be
        ext_cycle = [curr_node]
        iter_count = 0
        while next_node.key != cycle_root.key:
            ext_cycle.append(next_node)
            next_node = PolygonDirectedGraph.next_exterior_node(next_node)
            if not next_node:
                return None  # we have hit a dead end in the cycle
            iter_count += 1
            if iter_count > max_iter:
                break  # we have gotten stuck in a loop

        return ext_cycle

    def exterior_cycles(self):
        """Get a list of lists where each sub-list is an exterior cycle of Nodes."""
        exterior_poly_lst = []  # list to store cycles
        explored_nodes = set()  # set to note explored exterior nodes
        max_iter = self.node_count + 1  # maximum length a cycle can be

        # loop through all of the nodes of the graph and find cycles
        for root_node in self.ordered_nodes:
            # make a note that the current node has been explored
            explored_nodes.add(root_node.key)  # mark the node as explored
            # get next exterior adjacent node and check that it's valid
            next_node = self.next_exterior_node(root_node)  # mark the node as explored
            is_valid = (next_node is not None) and (next_node.key not in explored_nodes)
            if not is_valid:
                continue
            # make a note that the next node has been explored
            explored_nodes.add(next_node.key)

            # traverse the loop of points until we get back to start or hit a dead end
            exterior_poly = [root_node]
            prev_node = root_node
            iter_count = 0
            while next_node.key != root_node.key:
                exterior_poly.append(next_node)
                explored_nodes.add(next_node.key)  # mark the node as explored
                follow_node = self.next_exterior_node_no_backtrack(
                    next_node, prev_node, explored_nodes)
                prev_node = next_node  # set as the previous node for the next step
                next_node = follow_node
                if next_node is None:
                    break  # we have hit a dead end in the cycle
                iter_count += 1
                if iter_count > max_iter:
                    print('Extraction of core polygons hit an endless loop.')
                    break  # we have gotten stuck in a loop
            exterior_poly_lst.append(exterior_poly)

        # return all of the exterior loops that were found
        return exterior_poly_lst

    @staticmethod
    def next_exterior_node_no_backtrack(node, previous_node, explored_nodes):
        """Get the next exterior node adjacent to the input node.

        This method is similar to the next_exterior_node method but it includes
        extra checks to handle intersections with 3 or more segments in the
        graph exterior cycles. In these cases a set of previously explored_nodes
        is used to ensure that no back-tracking happens over the search of the
        network, which can lead to infinite looping through the graph. Furthermore,
        the previous_node is used to select the pathway with the smallest angle
        difference with the previous direction. This leads the result towards
        minimal polygons with fewer self-intersecting loops.

        Args:
            node: A _Node object for which the next node will be returned.
            previous_node: A _Node object for the node that came before
                the current one in the loop. This will be used in the event that
                multiple exterior nodes are found connecting to the input node.
                In this case, the exterior node with the smallest angle difference
                with the previous direction will be returned. This leads the
                result towards minimal polygons and away from self-intersecting
                exterior loops like a bowtie.

        Returns:
            Next node that defines exterior edge, or None if all adjacencies are
            bidirectional.
        """
        # loop through the all adjacent nodes and determine if they are exterior
        next_nodes = []
        for _next_node in node.adj_lst:
            if _next_node.exterior:  # user has labeled it as exterior; we're done!
                return _next_node
            elif _next_node.exterior is None:  # don't know if it's interior or exterior
                # if user-assigned attribute isn't defined, check bi-directionality
                if not PolygonDirectedGraph.is_edge_bidirect(node, _next_node):
                    next_nodes.append(_next_node)

        # evaluate whether there is one obvious choice for the next node
        if len(next_nodes) <= 1:
            return next_nodes[0] if len(next_nodes) == 1 else None
        next_nodes = [nn for nn in next_nodes if nn.key not in explored_nodes]
        if len(next_nodes) <= 1:
            return next_nodes[0] if len(next_nodes) == 1 else None

        # if we have multiple exterior nodes, use the previous node to find the best one
        prev_dir = previous_node.pt - node.pt  # yields a vector
        next_angles = []
        for next_node in next_nodes:
            edge_dir = next_node.pt - node.pt  # yields a vector
            next_angles.append(prev_dir.angle(edge_dir * -1))
        sorted_nodes = [n for _, n in sorted(zip(next_angles, next_nodes),
                                             key=lambda pair: pair[0])]
        return sorted_nodes[0]  # return the node making the smallest angle

    @staticmethod
    def next_exterior_node(node):
        """Get the next exterior node adjacent to consumed node.

        If there are adjacent nodes that are labeled as exterior, with True or
        False defining the _Node.exterior property, the first of such nodes in
        the adjacency list will be returned as the next one. Otherwise, the
        bi-directionality will be used to determine whether the next node is
        exterior.

        Args:
            node: A _Node object for which the next node will be returned.

        Returns:
            Next node that defines exterior edge, or None if all adjacencies are
            bidirectional.
        """
        # loop through the adjacency and find an exterior node
        for _next_node in node.adj_lst:
            if _next_node.exterior:  # user has labeled it as exterior; we're done!
                return _next_node
            elif _next_node.exterior is None:  # don't know if it's interior or exterior
                # if user-assigned attribute isn't defined, check bi-directionality
                if not PolygonDirectedGraph.is_edge_bidirect(node, _next_node):
                    return _next_node
        return None

    @staticmethod
    def is_edge_bidirect(node1, node2):
        """Are two nodes bidirectional.

        Args:
            node1: _Node object
            node2: _Node object

        Returns:
            True if node1 and node2 are in each other's adjacency list,
            else False.
        """
        return node1.key in (n.key for n in node2.adj_lst) and \
            node2.key in (n.key for n in node1.adj_lst)

    def __repr__(self):
        """Represent PolygonDirectedGraph."""
        s = ''
        for n in self.ordered_nodes:
            s += '{}, [{}]\n'.format(
                n.pt.to_array(),
                ', '.join([str(_n.pt.to_array()) for _n in n.adj_lst]))
        return s


def skeleton_as_directed_graph(boundary, holes=None, tolerance=1e-5):
    """Compute the straight skeleton of a shape as a PolygonDirectedGraph.

    Args:
        boundary: A ladybug-geometry Polygon2D for the boundary around the shape
            for which the straight skeleton will be computed.
        holes: An optional list of ladybug-geometry Polygon2D for the holes within
            the shape for which a straight skeleton will be computed. If None,
            it will be assumed that no holes exist in the shape. (Default: None).
        tolerance: Tolerance for point equivalence. (Default: 1e-5).

    Returns:
        A PolygonDirectedGraph object that represents the network and boundary
        of the straight skeleton. All interior connections in the graph are
        assumed to be bi-directional. The edges (including the boundary
        and holes) are uni-directional with the outer boundary being counterclockwise
        and the holes being clockwise. In other words, the fill of the shape is
        always to the left of each exterior edge. The nodes at the boundary
        and the holes have the exterior property set to True.
    """
    # get the segments representing the straight skeleton
    skeleton = skeleton_as_edge_list(boundary, holes, tolerance)
    skeleton, is_intersect_topology = _intersect_skeleton_segments(skeleton, tolerance)
    skeleton = _remove_segments_outside_boundary(skeleton, boundary, tolerance)

    # ensure the boundary and holes are oriented correctly for the graph
    if boundary.is_clockwise:
        boundary = boundary.reverse()
    loops = [boundary.vertices]
    if holes is not None:
        for hole in holes:
            if hole.is_clockwise:
                loops.append(hole.vertices)
            else:
                loops.append(hole.reverse().vertices)

    # make the directed graph and add the nodes for the boundary + holes
    dg_tol = 2 * tolerance  # use 2 times the tolerance to ensure hashing works
    dg = PolygonDirectedGraph(tol=dg_tol)
    for loop_count, vertices in enumerate(loops):
        for j in range(len(vertices) - 1):
            curr_v = vertices[j]
            next_v = vertices[j + 1]
            k = dg.add_node(curr_v, [next_v], exterior=True)
            if j == 0:
                if loop_count == 0:
                    dg.outer_root_key = k
                else:
                    dg.hole_root_keys.append(k)
        dg.add_node(vertices[-1], [vertices[0]], exterior=True)  # close loop

    # add the straight skelton to the graph
    for seg in skeleton:
        # add a bidirectional edge to represent skeleton edges
        dg.add_node(seg.p2, [seg.p1])
        dg.add_node(seg.p1, [seg.p2], exterior=False)

    # set the property to track whether the graph is self-intersecting
    dg.is_intersect_topology = is_intersect_topology
    return dg


def skeleton_as_cycle_polygons(boundary, holes=None, tolerance=1e-5):
    """Compute the straight skeleton of a shape as a list of cycle polygons.

    Args:
        boundary: A ladybug-geometry Polygon2D for the boundary around the shape
            for which the straight skeleton will be computed.
        holes: An optional list of ladybug-geometry Polygon2D for the holes within
            the shape for which a straight skeleton will be computed. If None,
            it will be assumed that no holes exist in the shape. (Default: None).
        tolerance: Tolerance for point equivalence. (Default: 1e-5).

    Returns:
        A list of Polygon2D that represent the straight skeleton of the input shape.
        There will be one Polygon2D for each edge of the shape (including both
        the boundary and the holes). Together, the Polygon2Ds will fill the
        entire input shape. There may be some overlap in between the output
        polygons if the straight skeleton did not have the correct topology.
    """
    # get the straight skeleton as a PolygonDirectedGraph
    dg = skeleton_as_directed_graph(boundary, holes, tolerance)

    # function to add cycles to the list of polygons to be returned
    cycle_polys = []

    def _add_cycle_polygon(min_cycle):
        """Add a cycle of Nodes to the list of polygons."""
        if min_cycle is not None:
            cycle_poly = Polygon2D([node.pt for node in min_cycle])
            cycle_polys.append(cycle_poly)

    # convert the edge cycles to Polygon2D
    exter_cycle = dg.exterior_cycle(dg.outer_root_node)
    for i, base_node in enumerate(exter_cycle[:-1]):
        next_node = exter_cycle[i + 1]
        _add_cycle_polygon(dg.min_cycle(next_node, base_node))
    _add_cycle_polygon(dg.min_cycle(exter_cycle[0], exter_cycle[-1]))
    if holes is not None:
        for hole_root in dg.hole_root_nodes:
            exter_cycle = dg.exterior_cycle(hole_root)
            for i, base_node in enumerate(exter_cycle[:-1]):
                next_node = exter_cycle[i + 1]
                _add_cycle_polygon(dg.min_cycle(next_node, base_node))
            _add_cycle_polygon(dg.min_cycle(exter_cycle[0], exter_cycle[-1]))
    return cycle_polys



# coding=utf-8
"""Implementation of the straight skeleton algorithm by Felkel and Obdrzalek[1].

The functions and classes here here are derived directly from the polyskel python
library by Armin Scipiades (@Bottfy), which is available at:
https://github.com/Botffy/polyskel

[1] Felkel, Petr and Stepan Obdrzalek. 1998. "Straight Skeleton Implementation." In
Proceedings of Spring Conference on Computer Graphics, Budmerice, Slovakia. 210 - 218.
"""
from __future__ import division

import heapq
from itertools import tee, islice, cycle, chain
from collections import namedtuple
import operator

# Geometry classes
from ladybug_geometry.geometry2d import Point2D, Ray2D, LineSegment2D, Polygon2D
from ladybug_geometry.intersection2d import intersect_line_segment2d, \
    intersect_line2d

# Polygon sorting classes
_OriginalEdge = namedtuple('_OriginalEdge', 'edge bisector_left, bisector_right')
Subtree = namedtuple('Subtree', 'source, height, sinks')
_SplitEventSubClass = namedtuple(
    '_SplitEvent', 'distance, intersection_point, vertex, opposite_edge')
_EdgeEventSubClass = namedtuple(
    '_EdgeEvent', 'distance intersection_point vertex_a vertex_b')


def _window(lst):
    """Window operator for lists.

    Consumes a list of items, and returns a zipped list of the previous,
    current and next items in the list, accessible by the same index.

    Args:
        lst: list

    Returns:
        Zipped list of previous, current and next items in list.
    """
    prevs, items, nexts = tee(lst, 3)
    prevs = islice(cycle(prevs), len(lst) - 1, None)
    nexts = islice(cycle(nexts), 1, None)
    return zip(prevs, items, nexts)


def _cross(a, b):
    """Get the determinant between two Vector2Ds and/or Point2Ds."""
    return a.x * b.y - b.x * a.y


def _approximately_equals(a, b):
    """Determine whether two Point2Ds or Vector2Ds are equal within a relative tolerance.
    """
    return a == b or (abs(a - b) <= max(abs(a), abs(b)) * 0.001)


def _normalize_contour(contour, tol):
    """Consumes list of x,y coordinate tuples and returns list of Point2Ds.

    Args:
        contour: list of x,y tuples from contour.
        tol: Number for point equivalence tolerance.

    Return:
         list of Point2Ds of contour.
    """
    contour = [Point2D(float(x), float(y)) for (x, y) in contour]
    normed_contour = []
    for prev, point, next in _window(contour):
        normed_prev = (point - prev).normalize()
        normed_next = (next - point).normalize()

        if not point.is_equivalent(next, tol) or \
                normed_prev.is_equivalent(normed_next, tol):
            normed_contour.append(point)

    return normed_contour


class _SplitEvent(_SplitEventSubClass):
    """A SplitEvent is a reflex vertex that splits an Edge Event.

    They therefore split the entire polygon and create new adjacencies between
    the split edge and each of the two edges incident to the reflex
    vertex (Felkel and Obdrzalek 1998, 1).
    """
    __slots__ = ()

    def __lt__(self, other):
        return self.distance < other.distance

    def __str__(self):
        return "{} Split event @ {} from {} to {}".format(
            self.distance, self.intersection_point, self.vertex, self.opposite_edge)


class _EdgeEvent(_EdgeEventSubClass):
    """An EdgeEvent is an edge extended from a perimeter edge, that shrinks to zero.

    This will make its neighboring edges adjacent (Felkel and Obdrzalek 1998, 2).
    """
    __slots__ = ()

    def __lt__(self, other):
        return self.distance < other.distance

    def __str__(self):
        return "{} Edge event @ {} between {} and {}".format(
            self.distance, self.intersection_point, self.vertex_a, self.vertex_b)


class _LAVertex:

    def __init__(self, point, edge_left, edge_right, direction_vectors=None, tol=1e-5):
        """A vertex in a double connected circular list of active vertices."""
        self.tol = tol
        self.point = point
        self.edge_left = edge_left
        self.edge_right = edge_right
        self.prev = None
        self.next = None
        self.lav = None
        # TODO this might be handled better. Maybe membership in lav implies validity?
        self._valid = True

        creator_vectors = (edge_left.v.normalize() * -1, edge_right.v.normalize())
        if direction_vectors is None:
            direction_vectors = creator_vectors

        self._is_reflex = (_cross(*direction_vectors)) < 0
        self._bisector = Ray2D(
            self.point, operator.add(*creator_vectors) * (-1 if self.is_reflex else 1))

    @property
    def bisector(self):
        return self._bisector

    @property
    def is_reflex(self):
        return self._is_reflex

    @property
    def original_edges(self):
        return self.lav._slav._original_edges

    def next_event(self):
        events = []
        if self.is_reflex:
            # A reflex vertex may generate a split event.
            # Split events happen when a vertex hits an opposite edge,
            # thereby splitting the polygon in two.
            for edge in self.original_edges:
                if edge.edge == self.edge_left or edge.edge == self.edge_right:
                    continue

                # A potential b is at the intersection between our own bisector and
                # the bisector of the angle between the tested edge and any one of our
                # own edges.
                # We choose 'less parallel' edge (to exclude a potentially parallel edge)

                # Make normalized copies of vectors
                norm_edge_left_v = self.edge_left.v.normalize()
                norm_edge_right_v = self.edge_right.v.normalize()
                norm_edge_v = edge.edge.v.normalize()

                # Compute dot
                leftdot = abs(norm_edge_left_v.dot(norm_edge_v))
                rightdot = abs(norm_edge_right_v.dot(norm_edge_v))
                selfedge = self.edge_left if leftdot < rightdot else self.edge_right

                # Make copies of edges and compute intersection
                self_copy = LineSegment2D(selfedge.p, selfedge.v)
                edge_copy = LineSegment2D(edge.edge.p, edge.edge.v)
                i = intersect_line_segment2d(edge_copy, self_copy)

                if i is not None and not _approximately_equals(i, self.point):
                    # locate candidate b
                    linvec = (self.point - i).normalize()
                    edvec = edge.edge.v.normalize()
                    if linvec.dot(edvec) < 0:
                        edvec = -edvec

                    bisecvec = edvec + linvec
                    if abs(bisecvec) == 0:
                        continue
                    bisector = LineSegment2D(i, bisecvec)
                    b = intersect_line2d(self.bisector, bisector)

                    if b is None:
                        continue

                    # Check eligibility of b.
                    # A valid b should lie within the area limited by the edge
                    # and the bisectors of its two vertices.
                    _left_bisector_norm = edge.bisector_left.v.normalize()
                    _left_b_norm = (b - edge.bisector_left.p).normalize()
                    xleft = _left_bisector_norm.determinant(_left_b_norm) > -self.tol

                    _right_bisector_norm = edge.bisector_right.v.normalize()
                    _right_b_norm = (b - edge.bisector_right.p).normalize()
                    xright = _right_bisector_norm.determinant(_right_b_norm) < self.tol

                    _edge_edge_norm = edge.edge.v.normalize()
                    _b_to_edge_norm = (b - edge.edge.p).normalize()
                    xedge = _edge_edge_norm.determinant(_b_to_edge_norm) < self.tol

                    if not (xleft and xright and xedge):
                        continue  # discarded candidate
                    # found valid candidate
                    _d2b = LineSegment2D(edge.edge.p, edge.edge.v).distance_to_point(b)
                    _new_split_event = _SplitEvent(_d2b, b, self, edge.edge)
                    events.append(_new_split_event)

        i_prev = self.bisector.intersect_line_ray(self.prev.bisector)
        i_next = self.bisector.intersect_line_ray(self.next.bisector)

        if i_prev is not None:
            left_seg = LineSegment2D(self.edge_left.p, self.edge_left.v)
            dist_to_i_prev = left_seg.distance_to_point(i_prev)
            events.append(_EdgeEvent(dist_to_i_prev, i_prev, self.prev, self))
        if i_next is not None:
            right_seg = LineSegment2D(self.edge_right.p, self.edge_right.v)
            dist_to_i_next = right_seg.distance_to_point(i_next)
            events.append(_EdgeEvent(dist_to_i_next, i_next, self, self.next))

        if not events:
            return None

        ev = min(events, key=lambda event:
                 self.point.distance_to_point(event.intersection_point))
        return ev

    def invalidate(self):
        if self.lav is not None:
            self.lav.invalidate(self)
        else:
            self._valid = False

    @property
    def is_valid(self):
        return self._valid

    def __str__(self):
        return "Vertex ({:.2f};{:.2f})".format(self.point.x, self.point.y)

    def __repr__(self):
        return 'Vertex ({}) ({:.2f};{:.2f}), bisector {}, edges {} {}'.format(
            'reflex' if self.is_reflex else 'convex',
            self.point.x, self.point.y, self.bisector,
            self.edge_left, self.edge_right)


class _SLAV:

    def __init__(self, polygon, tol=1e-5):
        """ A set of circular lists of active vertices.

        It stores a loop of vertices for the polygon created during the straight
        skeleton computation (Felkel and Obdrzalek 1998).
        """
        self.tol = tol
        contours = [_normalize_contour(polygon, tol)]
        self._lavs = [_LAV.from_polygon(contour, self) for contour in contours]

        # store original polygon edges for calculating split events
        self._original_edges = [
            _OriginalEdge(
                LineSegment2D(vertex.prev.point, vertex.point),
                vertex.prev.bisector, vertex.bisector
            ) for vertex in chain.from_iterable(self._lavs)
        ]

    def __iter__(self):
        for lav in self._lavs:
            yield lav

    def __len__(self):
        return len(self._lavs)

    def empty(self):
        return len(self._lavs) == 0

    def handle_edge_event(self, event):
        """Resolves adjacency of new edge event.

        This function resolves the edge event with previous edges, LAV, and then stores
        edge information in a Subtree.

        Args:
            event: EdgeEvent

        Returns:
            Subtree namedTuple
        """
        sinks = []
        events = []

        lav = event.vertex_a.lav
        # triangle; one sink point
        if event.vertex_a.prev == event.vertex_b.next:
            self._lavs.remove(lav)
            for vertex in list(lav):
                sinks.append(vertex.point)
                vertex.invalidate()
        else:
            new_vertex = lav.unify(event.vertex_a, event.vertex_b,
                                   event.intersection_point)
            if lav.head in (event.vertex_a, event.vertex_b):
                lav.head = new_vertex
            sinks.extend((event.vertex_a.point, event.vertex_b.point))
            next_event = new_vertex.next_event()
            if next_event is not None:
                events.append(next_event)

        return (Subtree(event.intersection_point, event.distance, sinks), events)

    def handle_split_event(self, event):
        """Consumes a split event.

        This function resolves the adjacency of new split event with previous edges, LAV,
        and then stores edge information in a Subtree.

        Args:
            event: EdgeEvent

        Returns:
            Subtree namedTuple
        """
        lav = event.vertex.lav
        sinks = [event.vertex.point]
        vertices = []
        x = None  # right vertex
        y = None  # left vertex
        norm = event.opposite_edge.v.normalize()
        for v in chain.from_iterable(self._lavs):
            if norm == v.edge_left.v.normalize() and \
                    event.opposite_edge.p == v.edge_left.p:
                x = v
                y = x.prev
            elif norm == v.edge_right.v.normalize() and \
                    event.opposite_edge.p == v.edge_right.p:
                y = v
                x = y.next

            if x:
                xleft = y.bisector.v.normalize().determinant(
                    (event.intersection_point - y.point).normalize()) >= -self.tol

                xright = x.bisector.v.normalize().determinant(
                    (event.intersection_point - x.point).normalize()) <= self.tol

                if xleft and xright:
                    break
                else:
                    x = None
                    y = None

        if x is None:
            return (None, [])

        v1 = _LAVertex(event.intersection_point, event.vertex.edge_left,
                       event.opposite_edge, tol=self.tol)
        v2 = _LAVertex(event.intersection_point, event.opposite_edge,
                       event.vertex.edge_right, tol=self.tol)

        v1.prev = event.vertex.prev
        v1.next = x
        event.vertex.prev.next = v1
        x.prev = v1

        v2.prev = y
        v2.next = event.vertex.next
        event.vertex.next.prev = v2
        y.next = v2

        new_lavs = None
        self._lavs.remove(lav)
        if lav != x.lav:
            # the split event actually merges two lavs
            self._lavs.remove(x.lav)
            new_lavs = [_LAV.from_chain(v1, self)]
        else:
            new_lavs = [_LAV.from_chain(v1, self), _LAV.from_chain(v2, self)]

        for lv in new_lavs:
            if len(lv) > 2:
                self._lavs.append(lv)
                vertices.append(lv.head)
            else:
                sinks.append(lv.head.next.point)
                for v in list(lv):
                    v.invalidate()

        events = []
        for vertex in vertices:
            next_event = vertex.next_event()
            if next_event is not None:
                events.append(next_event)

        event.vertex.invalidate()
        return (Subtree(event.intersection_point, event.distance, sinks), events)


class _LAV:

    def __init__(self, slav):
        """A single circular list of active vertices.

        Stored in a SLAV (Felkel and Obdrzalek 1998, 2).
        """
        self.head = None
        self._slav = slav
        self._len = 0
        self.tol = slav.tol

    @classmethod
    def from_polygon(cls, polygon, slav):
        """Construct a LAV from a list of point coordinates representing a polygon.

        Args:
            polygon: list of points (tuple of x,y coordinates).
            slav: SLAV (a set of circular lists of active vertices).

        Returns:
            LAV (single circular list of active vertices).
        """
        lav = cls(slav)
        for prev, point, next in _window(polygon):
            lav._len += 1
            vertex = _LAVertex(
                point,
                LineSegment2D.from_end_points(prev, point),
                LineSegment2D.from_end_points(point, next),
                tol=slav.tol
            )
            vertex.lav = lav
            if lav.head is None:
                lav.head = vertex
                vertex.prev = vertex.next = vertex
            else:
                vertex.next = lav.head
                vertex.prev = lav.head.prev
                vertex.prev.next = vertex
                lav.head.prev = vertex
        return lav

    @classmethod
    def from_chain(cls, head, slav):
        """Creates new LAV from consumed _LAVertex, and reference _SLAV.

        Args:
            head: Head _LAVertex that creates new _LAV loop.
            slav: Reference SLAV (a set of circular lists of active vertices).

        Returns:
            LAV (a single circular list of active vertices).
        """
        lav = cls(slav)
        lav.head = head
        for vertex in lav:
            lav._len += 1
            vertex.lav = lav
        return lav

    def invalidate(self, vertex):
        """Sets vertex validity to False, and handles head relationship.

        Args:
            vertex: _LAVertex to be invalidated.
        """
        assert vertex.lav is self, 'Tried to invalidate a vertex that is not mine'
        vertex._valid = False
        if self.head == vertex:
            self.head = self.head.next
        vertex.lav = None

    def unify(self, vertex_a, vertex_b, point):
        """Generate new _LAVertex from input Point2D and resolve adjacency to old edges.

        Args:
            vertex_a: _LAVertex
            vertex_b: _LAVertex
            point: Intersection point from angle bisectors from vertex_a and
                vertex_b as a Point2D.

        Returns:
            _LAVertex of intersection point.
        """
        replacement = _LAVertex(
            point, vertex_a.edge_left, vertex_b.edge_right,
            (vertex_b.bisector.v.normalize(), vertex_a.bisector.v.normalize()),
            tol=vertex_a.tol
        )
        replacement.lav = self

        if self.head in [vertex_a, vertex_b]:
            self.head = replacement

        vertex_a.prev.next = replacement
        vertex_b.next.prev = replacement
        replacement.prev = vertex_a.prev
        replacement.next = vertex_b.next

        vertex_a.invalidate()
        vertex_b.invalidate()

        self._len -= 1
        return replacement

    def __str__(self):
        return "LAV {}".format(id(self))

    def __repr__(self):
        return "{} = {}".format(str(self), [vertex for vertex in self])

    def __len__(self):
        return self._len

    def __iter__(self):
        cur = self.head
        while True:
            yield cur
            cur = cur.next
            if cur == self.head:
                return

    def _show(self):
        cur = self.head
        while True:
            print(cur.__repr__())
            cur = cur.next
            if cur == self.head:
                break


class _EventQueue:

    def __init__(self):
        """A priority queue that stores vertices of the polygon."""
        self.__data = []

    def put(self, item):
        if item is not None:
            heapq.heappush(self.__data, item)

    def put_all(self, iterable):
        for item in iterable:
            heapq.heappush(self.__data, item)

    def get(self):
        return heapq.heappop(self.__data)

    def empty(self):
        return len(self.__data) == 0

    def peek(self):
        return self.__data[0]

    def show(self):
        for item in self.__data:
            print(item)


def _merge_sources(skeleton):
    """Merge the sources of a straight skeleton.

    In highly symmetrical shapes with reflex vertices, multiple sources may share
    the same  location. This function merges those sources.
    """
    sources = {}
    to_remove = []
    for i, p in enumerate(skeleton):
        source = tuple(i for i in p.source)
        if source in sources:
            source_index = sources[source]
            # source exists, merge sinks
            for sink in p.sinks:
                if sink not in skeleton[source_index].sinks:
                    skeleton[source_index].sinks.append(sink)
            to_remove.append(i)
        else:
            sources[source] = i
    for i in reversed(to_remove):
        skeleton.pop(i)


def _clean_sinks(skeleton, tol=1e-5):
    """Remove cases where the source and sink point are the same.

    This can occur as an unintended effect of merging sources.
    """
    clean_subtrees = []
    for s_tree in skeleton:
        source, clean_sinks, is_invalid = s_tree.source, [], False
        for sink in s_tree.sinks:
            if source.is_equivalent(sink, tol):
                is_invalid = True
            else:
                clean_sinks.append(sink)
        if is_invalid:  # create a new Subtree
            clean_subtrees.append(Subtree(source, s_tree.height, clean_sinks))
        else:
            clean_subtrees.append(s_tree)
    return clean_subtrees


def _skeletonize(polygon, tol=1e-5):
    """Compute the straight skeleton of a Polygon.

    Args:
        polygon: A list of 2D vertices in clockwise order (when viewed from
            above the XY plane). For example, a square can be represented with
            the following: [[0,1], [1,1], [0,0], [1,0]]
        tol: Tolerance for point equivalence. (Default: 1e-5).

    Returns:
        The straight skeleton as a list of "subtrees", which are in the form of
        (source, height, sinks). Source is the highest points, height is its
        height, and sinks are the point connected to the source.
    """
    slav = _SLAV(polygon, tol=tol)
    output = []
    prioque = _EventQueue()

    for lav in slav:
        for vertex in lav:
            v = vertex.next_event()
            prioque.put(v)

    # While the priority queue or SLAV is not empty, compute the intersection of
    # bisectors at edge. The 'handle_edge_event' method computes the next event
    # from the new intersection via the next_event method. Thus, this while loop
    # iteratively adds new events without recursion.
    while not (prioque.empty() or slav.empty()):
        i = prioque.get()  # vertex a, b is self or next vertex
        # Handle edge or split events.
        # arc: subtree(event.intersection_point, event.distance, sinks)
        # events: updated events with new vertex
        if isinstance(i, _EdgeEvent):
            if not i.vertex_a.is_valid or not i.vertex_b.is_valid:
                continue
            (arc, events) = slav.handle_edge_event(i)
        elif isinstance(i, _SplitEvent):
            if not i.vertex.is_valid:
                continue
            (arc, events) = slav.handle_split_event(i)
        prioque.put_all(events)

        # As we traverse priorque, output list of "subtrees", which are in the form
        # of (source, height, sinks) where source is the highest points, height is
        # its distance to an edge, and sinks are the point connected to the source.
        if arc is not None:
            output.append(arc)
    # merge the sources and clean the result
    _merge_sources(output)
    output = _clean_sinks(output, tol=1e-5)
    return output


def _intersect_skeleton_segments(skeleton, tolerance=1e-5):
    """Intersect all of the LineSegment2D of the skeleton and split them.

    Args:
        skeleton: A list of ladybug-geometry LineSegment2D for the segments
            of the straight skeleton.
        tolerance: The tolerance at which the intersection will be computed.

    Returns:
        A tuple with two items.

        * split_skeleton -- A list of LineSegment2D objects for the skeleton
            split through self-intersection.

        * is_intersect_topology -- A boolean value for whether the topology of the
            input skeleton is self-intersecting. This value is True whenever
            there are segments that were split as part of this operation.
    """
    # extend skeleton segments a little to ensure intersections happen
    under_tol = tolerance * 0.99
    ext_skeleton = []
    for seg in skeleton:
        m_v = seg.v.normalize() * under_tol
        ext_seg = LineSegment2D.from_end_points(seg.p1.move(-m_v), seg.p2.move(m_v))
        ext_skeleton.append(ext_seg)

    # compute all of the intersection points across the skeleton
    intersect_pts = [[] for _ in skeleton]
    for i, seg in enumerate(skeleton):
        try:
            for other_seg in ext_skeleton[:i] + ext_skeleton[i + 1:]:
                int_pt = intersect_line_segment2d(seg, other_seg)
                if int_pt is None or int_pt.is_equivalent(seg.p1, tolerance) or \
                        int_pt.is_equivalent(seg.p2, tolerance):
                    continue
                # we have found an intersection point where segments should be split
                intersect_pts[i].append(int_pt)
        except IndexError:
            pass  # we have reached the end of the list

    # loop through the segments and split them at the intersection points
    split_skeleton, is_intersect_topology = [], False
    for seg, split_pts in zip(skeleton, intersect_pts):
        if len(split_pts) == 0:
            split_skeleton.append(seg)
        elif len(split_pts) == 1:  # split the segment in two
            is_intersect_topology = True
            int_pt = split_pts[0]
            split_skeleton.append(LineSegment2D.from_end_points(seg.p1, int_pt))
            split_skeleton.append(LineSegment2D.from_end_points(int_pt, seg.p2))
        else:  # sort the points along the segment to split it
            is_intersect_topology = True
            pt_dists = [seg.p1.distance_to_point(ipt) for ipt in split_pts]
            sort_obj = sorted(zip(pt_dists, split_pts), key=lambda pair: pair[0])
            sort_pts = [x for _, x in sort_obj]
            sort_pts.append(seg.p2)
            pr_pt = seg.p1
            for s_pt in sort_pts:
                if not pr_pt.is_equivalent(s_pt, tolerance):
                    split_skeleton.append(LineSegment2D.from_end_points(pr_pt, s_pt))
                pr_pt = s_pt

    return split_skeleton, is_intersect_topology


def _remove_segments_outside_boundary(skeleton, boundary, tolerance=1e-5):
    """Remove LineSegment2D that are outside the boundary of the parent shape.

    This can be used to clean up the result after intersection of segments
    since the skeleton routines can sometimes fail to produce a result that
    is completely within the parent shape.

    Args:
        skeleton: A list of ladybug-geometry LineSegment2D for the segments
            of the straight skeleton.
        boundary: A Polygon2D for the boundary of the shape. Segments that lie
            outside of this boundary beyond the tolerance will be removed from
            the result.
        tolerance: The tolerance for distinguishing whether skeleton points lie
            outside the boundary.

    Returns:
        A tuple with two items.

        * clean_skeleton -- A list of LineSegment2D objects with segments removed
            that outside of the boundary.
    """
    clean_skeleton = []
    for seg in skeleton:
        p1, p2 = seg.p1, seg.p2
        if boundary.point_relationship(p2, tolerance) >= 0 and \
                boundary.point_relationship(p1, tolerance) >= 0:
            clean_skeleton.append(seg)
    return clean_skeleton


def skeleton_as_subtree_list(boundary, holes=None, tolerance=1e-5):
    """Get a straight skeleton as a list of Subtree source and sink points.

    Args:
        boundary: A ladybug-geometry Polygon2D for the boundary around the shape
            for which the straight skeleton will be computed.
        holes: An optional list of ladybug-geometry Polygon2D for the holes within
            the shape for which a straight skeleton will be computed. If None,
            it will be assumed that no holes exist in the shape. (Default: None).
        tolerance: Tolerance for point equivalence. (Default: 1e-5).

    Returns:
        A list of Subtree objects that represent the straight skeleton.
    """
    # merge the holes into the boundary if they are specified
    if holes is not None and len(holes) != 0:
        bound_pts = list(boundary.vertices)
        hole_pts = [list(h.vertices) for h in holes]
        boundary = Polygon2D.from_shape_with_holes(bound_pts, hole_pts)

    # pre-process the boundary to be sure it is in the right order
    boundary = boundary.remove_duplicate_vertices(tolerance)
    if not boundary.is_clockwise:
        boundary = boundary.reverse()

    # skeletonize the objects
    skel_tol = tolerance / 100  # use a finer tolerance for actual skeleton
    subtree_list = _skeletonize(boundary.to_array(), skel_tol)

    # if holes were merged into the boundary, try to clean up the merge seams
    if holes is not None and len(holes) != 0:
        # gather the seams along which holes were merged
        all_segs, seam_segs = boundary.segments, []
        for i, seg in enumerate(all_segs):
            try:
                for o_seg in all_segs[i + 1:]:
                    if seg.p1.is_equivalent(o_seg.p2, tolerance) and \
                            seg.p2.is_equivalent(o_seg.p1, tolerance):
                        seam_segs.append(seg)
                        break
            except IndexError:
                pass  # we have reached the end of the list
        # replace the subtrees along the seam with the seam itself
        for seam in seam_segs:
            s_p1, s_p2 = seam.p1, seam.p2
            new_subtrees, connecting_pts, new_height = [], [], 0
            for sub_tree in subtree_list:
                sink_pts = sub_tree.sinks
                replace_sub = False
                if len(sink_pts) == 2:
                    if s_p1.is_equivalent(sink_pts[0], tolerance) or \
                            s_p1.is_equivalent(sink_pts[1], tolerance):
                        if s_p2.is_equivalent(sink_pts[0], tolerance) or \
                                s_p2.is_equivalent(sink_pts[1], tolerance):
                            new_height = sub_tree.height
                            replace_sub = True
                            connecting_pts.append(sub_tree.source)
                if not replace_sub:
                    new_subtrees.append(sub_tree)
            subtree_list = new_subtrees
            seam_sub = Subtree(seam.midpoint, new_height, connecting_pts + [s_p1, s_p2])
            subtree_list.append(seam_sub)

    return subtree_list


def skeleton_as_edge_list(boundary, holes=None, tolerance=1e-5, intersect=False):
    """Get a straight skeleton as list of LineSegment2D.

    Args:
        boundary: A ladybug-geometry Polygon2D for the boundary around the shape
            for which the straight skeleton will be computed.
        holes: An optional list of ladybug-geometry Polygon2D for the holes within
            the shape for which a straight skeleton will be computed. If None,
            it will be assumed that no holes exist in the shape. (Default: None).
        tolerance: Tolerance for point equivalence. (Default: 1e-5).
        intersect: A boolean to note whether the segments of the skeleton should
            be intersected with one another before being returned. This can
            help make the skeleton more usable in the event that its topology
            is not correct. (Default: False).

    Returns:
        A list of LineSegment2D that represent the straight skeleton.
    """
    # get the Subtree representation of the straight skeleton
    subtree_list = skeleton_as_subtree_list(boundary, holes, tolerance)

    # extract the LineSegment2D from the Subtree representation
    skeleton = []
    for subtree in subtree_list:
        source_pt = subtree.source
        for sink_pt in subtree.sinks:
            edge_seg = LineSegment2D.from_end_points(
                Point2D(source_pt.x, source_pt.y), Point2D(sink_pt.x, sink_pt.y))
            skeleton.append(edge_seg)

    if intersect:  # intersect skeleton segments to split them
        skeleton, _ = _intersect_skeleton_segments(skeleton, tolerance)
        skeleton = _remove_segments_outside_boundary(skeleton, boundary, tolerance)
    return skeleton



# coding=utf-8
"""Test the class for handling pathways through the straight skeleton network."""
from __future__ import division
import pytest

from ladybug_geometry_polyskel.polygraph import PolygonDirectedGraph, \
    skeleton_as_directed_graph, skeleton_as_cycle_polygons, _vector2hash
from ladybug_geometry.geometry2d.polygon import Polygon2D
from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D


def _cmpstr(item1, item2):
    return '{} vs {}'.format(item1, item2)


def test_vector2hash():
    """Test the vector hash method"""
    # Integer vector
    vec = Vector2D(1, 1)
    hash_val = _vector2hash(vec, tol=0)
    assert hash_val == '(1.0, 1.0)'

    # Float with 1e-4 points
    vec = Vector2D(1.1111, 1.1111)
    hash_val = _vector2hash(vec, tol=1e-4)
    assert hash_val == '(1.1111, 1.1111)'

    # Float with 1e-4 points w/ rounding
    vec = Vector2D(1.11116, 1.11116)
    hash_val = _vector2hash(vec, tol=1e-4)
    assert hash_val == '(1.1112, 1.1112)'

    # Round to the ones
    vec = Vector2D(115.11116, 115.11116)
    hash_val = _vector2hash(vec, tol=1)
    assert hash_val == '(115.0, 115.0)'

    # Round to the tenths
    vec = Vector2D(116.11116, 116.11116)
    hash_val = _vector2hash(vec, tol=10)
    assert hash_val == '(120.0, 120.0)'

    # 10 digit tolerance w/o rounding
    vec = Vector2D(1.0123456789, 1.0123456789)
    hash_val = _vector2hash(vec, tol=1e-10)
    assert hash_val == '(1.0123456789, 1.0123456789)'

    # 10 digit tolerance w/ rounding
    vec = Vector2D(1.01234567888, 1.01234567888)
    hash_val = _vector2hash(vec, tol=1e-10)
    assert hash_val == '(1.0123456789, 1.0123456789)'

    # 7 digit tolerance w/ rounding
    vec = Vector2D(1.01234567888, 1.01234567888)
    hash_val = _vector2hash(vec, tol=1e-7)
    assert hash_val == '(1.0123457, 1.0123457)'


def test_vector2hash_non_unity():
    """Test the vector hash method with tolerance that don't have a base of 1."""
    vec = Vector2D(1.115, 1.115)
    hash_val = _vector2hash(vec, tol=0.001)
    assert hash_val == '(1.115, 1.115)'

    vec = Vector2D(1.115, 1.115)
    hash_val = _vector2hash(vec, tol=0.002)
    assert hash_val == '(1.114, 1.114)'

    hash_val = _vector2hash(vec, tol=0.003)
    assert hash_val == '(1.116, 1.116)'

    hash_val = _vector2hash(vec, tol=0.004)
    assert hash_val == '(1.116, 1.116)'

    hash_val = _vector2hash(vec, tol=0.005)
    assert hash_val == '(1.115, 1.115)'

    hash_val = _vector2hash(vec, tol=0.006)
    assert hash_val == '(1.116, 1.116)'

    vec = Vector2D(1.1156, 1.1157)
    hash_val = _vector2hash(vec, tol=0.002)
    assert hash_val == '(1.116, 1.116)'


def test_dg_noskel():
    """Test the dg with no skeleton"""
    # Points
    pt_array = [[0, 0], [6, 0], [6, 6], [3, 9], [0, 6]]

    # Make the polygon
    polygon = Polygon2D.from_array(pt_array)

    # Make the check cases
    chk_pt_lst = [Point2D.from_array(p) for p in pt_array]

    # Inititalize a dg object
    d = PolygonDirectedGraph(1e-5)
    vertices = polygon.vertices

    # Add edges to dg
    for i in range(len(vertices) - 1):
        k = d.add_node(vertices[i], [vertices[i + 1]])
        if i == 0:
            d.outer_root_key = k
    d.add_node(vertices[-1], [vertices[0]])

    # Test number
    assert len(chk_pt_lst) == d.node_count, _cmpstr(len(chk_pt_lst), d.node_count)

    # Test root
    assert d.node(d.outer_root_key)._order == 0

    # Test adjacencies are correct
    curr_node = d.node(d.outer_root_key)
    for chk_pt in chk_pt_lst:
        assert chk_pt.is_equivalent(curr_node.pt, 1e-5), _cmpstr(chk_pt, curr_node.pt)

        # Increment
        curr_node = curr_node.adj_lst[0]

    # Test the adj matrix
    amtx = d.adj_matrix()

    # Adj matrix to test against
    chk_amtx = [
         [0, 1, 0, 0, 0],  # 0
         [0, 0, 1, 0, 0],  # 1
         [0, 0, 0, 1, 0],  # 2
         [0, 0, 0, 0, 1],  # 3
         [1, 0, 0, 0, 0]]  # 4
    #     0, 1, 2, 3, 4

    # Test if the adj matrix is correct
    for i in range(len(chk_amtx)):
        for j in range(len(chk_amtx[0])):
            assert amtx[i][j] == chk_amtx[i][j], _cmpstr(amtx[i][j], chk_amtx[i][j])


def test_skeleton_as_directed_graph_triangle():
    """Test the skeleton_as_directed_graph function."""
    polygon_verts = [
        [0., 0.],
        [7., 0.],
        [4., 4.]
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    cycle_polys = skeleton_as_cycle_polygons(polygon, tolerance=1e-5)
    for poly in cycle_polys:
        assert isinstance(poly, Polygon2D)
        assert len(poly.vertices) == 3


def test_skeleton_as_directed_graph_square():
    """Test the skeleton_as_directed_graph function."""
    polygon_verts = [
        [0.,  0.],
        [10., 0.],
        [10., 10.],
        [0., 10.]
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    cycle_polys = skeleton_as_cycle_polygons(polygon, tolerance=1e-5)
    for poly in cycle_polys:
        assert isinstance(poly, Polygon2D)
        assert len(poly.vertices) == 3


def test_skeleton_as_directed_graph_one_hole():
    """Test skeleton_as_directed_graph using a shape with one hole."""
    polygon_verts = [
        [0., 0.],
        [3., 0.],
        [3., 3.],
        [0., 3.]
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    hole_verts = [
        [1., 1.],
        [2., 1.],
        [2., 2.],
        [1., 2.]
    ]
    hole = Polygon2D.from_array(hole_verts)

    dg = skeleton_as_directed_graph(polygon, [hole], tolerance=1e-5)
    ext_cycle = dg.exterior_cycle(dg.outer_root_node)
    min_cycle = dg.min_cycle(ext_cycle[1], ext_cycle[0])
    assert len(min_cycle) == 4
    assert min_cycle[-1].pt.x == pytest.approx(0., rel=1e-3)
    assert min_cycle[-1].pt.y == pytest.approx(0., rel=1e-3)
    assert min_cycle[0].pt.x == pytest.approx(3., rel=1e-3)
    assert min_cycle[0].pt.y == pytest.approx(0., rel=1e-3)
    assert not dg.is_intersect_topology

    cycle_polys = skeleton_as_cycle_polygons(polygon, [hole], tolerance=1e-5)
    for poly in cycle_polys:
        assert isinstance(poly, Polygon2D)
        assert len(poly.vertices) == 4 or len(poly.vertices) == 5


def test_skeleton_as_directed_graph_concave_two_holes():
    """Test concave with two holes."""
    polygon_verts = [
        [0.7, 0.2],
        [2, 0],
        [2, 2],
        [1, 1],
        [0, 2],
        [0, 0]
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    hole1_verts = [
        [0.6, 1.2],
        [1, 0.8],
        [1.5, 0.6],
        [0.6, 0.6]
    ]
    hole1 = Polygon2D.from_array(hole1_verts)

    hole2_verts = [
        [1.3, 0.5],
        [1.5, 0.25],
        [1.1, 0.25]
    ]
    hole2 = Polygon2D.from_array(hole2_verts)

    cycle_polys = skeleton_as_cycle_polygons(polygon, [hole1, hole2], tolerance=1e-5)
    for poly in cycle_polys:
        assert isinstance(poly, Polygon2D)
        assert 4 <= len(poly.vertices) <= 7


def test_skeleton_as_directed_graph_bad_topology():
    """Test a known case where polyskel returns bad topology"""
    polygon_verts = [
        [-3.10661836926, 21.3923815366],
        [-3.10661836926, 19.7597225480],
        [-7.59797976254, 19.7597225480],
        [-7.59797976254, 21.3923815366],
        [-6.59662650790, 21.3923815366],
        [-6.59662650790, 21.8240139474],
        [-4.11444320173, 21.8240139474],
        [-4.11444320173, 21.3923815366]
    ]
    polygon = Polygon2D.from_array(polygon_verts)
    cycle_polys = skeleton_as_cycle_polygons(polygon, tolerance=0.01)
    assert len(cycle_polys) == 8
    for poly in cycle_polys:
        assert isinstance(poly, Polygon2D)
        assert 3 <= len(poly.vertices) <= 7


def test_skeleton_as_directed_graph_bad_topology_v2():
    """Test another known case where polyskel returns bad topology"""
    polygon_verts = [
        [-3.69996219119, 17.0824478496],
        [-3.69996219119, 15.9029877055],
        [-8.19132358447, 15.9029877055],
        [-8.19132358447, 17.5356466941],
        [-7.18997032982, 17.5356466941],
        [-7.18997032982, 17.9672791049],
        [-4.70778702366, 17.9672791049],
        [-4.70778702366, 17.0824478496]
    ]
    polygon = Polygon2D.from_array(polygon_verts)
    cycle_polys = skeleton_as_cycle_polygons(polygon, tolerance=0.01)
    assert len(cycle_polys) == 8
    for poly in cycle_polys:
        assert isinstance(poly, Polygon2D)
        assert 3 <= len(poly.vertices) <= 8



# coding=utf-8
"""Test the methods for splitting polygons with a straight skeleton."""
from __future__ import division
import pytest

from ladybug_geometry.geometry2d import Polygon2D
from ladybug_geometry.geometry3d import LineSegment3D, Face3D

from ladybug_geometry_polyskel.polysplit import perimeter_core_subpolygons, \
    perimeter_core_subfaces, perimeter_core_subfaces_and_skeleton


def test_perimeter_core_subpolygons_triangle():
    """Test the perimeter_core_subpolygons function with a triangle."""
    polygon_verts = [
        [0., 0.],
        [8., 0.],
        [4., 4.]
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    perim_polys, core_polys = perimeter_core_subpolygons(polygon, 1.0, tolerance=1e-5)
    assert len(perim_polys) == 3
    for poly in perim_polys:
        assert len(poly) == 4
    assert perim_polys[0].area == pytest.approx(5.58578643, rel=1e-3)
    assert perim_polys[1].area == pytest.approx(3.949747468, rel=1e-3)
    assert perim_polys[2].area == pytest.approx(3.949747468, rel=1e-3)
    assert len(core_polys) == 1
    assert len(core_polys[0]) == 3
    assert core_polys[0].area == pytest.approx(2.51471862, rel=1e-3)

    perim_polys, core_polys = perimeter_core_subpolygons(polygon, 6.0, tolerance=1e-5)
    assert len(perim_polys) == 3
    for poly in perim_polys:
        assert len(poly) == 3
    assert len(core_polys) == 0


def test_perimeter_core_subpolygons_square():
    """Test the perimeter_core_subpolygons function with a square."""
    polygon_verts = [
        [0.,  0.],
        [10., 0.],
        [10., 10.],
        [0., 10.]
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    perim_polys, core_polys = perimeter_core_subpolygons(polygon, 2.0, tolerance=1e-5)
    assert len(perim_polys) == 4
    for poly in perim_polys:
        assert len(poly) == 4
        assert poly.area == pytest.approx(16., rel=1e-3)
    assert len(core_polys) == 1
    assert len(core_polys[0]) == 4
    assert core_polys[0].area == pytest.approx(36., rel=1e-3)

    perim_polys, core_polys = perimeter_core_subpolygons(polygon, 6.0, tolerance=1e-5)
    assert len(perim_polys) == 4
    for poly in perim_polys:
        assert len(poly) == 3
        assert poly.area == pytest.approx(25., rel=1e-3)
    assert len(core_polys) == 0

    perim_polys, core_polys = perimeter_core_subpolygons(polygon, 5.0, tolerance=1e-5)
    assert len(perim_polys) == 4
    for poly in perim_polys:
        assert len(poly) == 3
        assert poly.area == pytest.approx(25., rel=1e-3)
    assert len(core_polys) == 0


def test_perimeter_core_subpolygons_one_hole():
    """Test perimeter_core_subpolygons using a shape with one hole."""
    polygon_verts = [
        [0., 0.],
        [3., 0.],
        [3., 3.],
        [0., 3.]
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    hole_verts = [
        [1., 1.],
        [2., 1.],
        [2., 2.],
        [1., 2.]
    ]
    hole = Polygon2D.from_array(hole_verts)

    perim_polys, core_polys = perimeter_core_subpolygons(
        polygon, 0.2, [hole], tolerance=1e-5, flat_core=False)
    assert len(perim_polys) == 8
    for poly in perim_polys:
        assert len(poly) == 4
        assert poly.area == pytest.approx(0.56, rel=1e-3) or \
            poly.area == pytest.approx(0.24, rel=1e-3)
    assert len(core_polys) == 1
    assert len(core_polys[0]) == 2
    assert core_polys[0][0].area == pytest.approx(6.76, rel=1e-3)
    assert core_polys[0][1].area == pytest.approx(1.96, rel=1e-3)

    perim_polys, core_polys = perimeter_core_subpolygons(
        polygon, 1.0, [hole], tolerance=1e-5, flat_core=False)
    assert len(perim_polys) == 8
    for poly in perim_polys:
        assert len(poly) == 4 or len(poly) == 5
        assert poly.area == pytest.approx(1.25, rel=1e-3) or \
            poly.area == pytest.approx(0.75, rel=1e-3)
    assert len(core_polys) == 0


def test_perimeter_core_subpolygons_two_holes():
    """Test perimeter_core_subpolygons with a concave geometry that has two holes."""
    polygon_verts = [
        [0.7, 0.2],
        [2, 0],
        [2, 2],
        [1, 1],
        [0, 2],
        [0, 0]
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    hole1_verts = [
        [0.6, 1.2],
        [1, 0.8],
        [1.5, 0.6],
        [0.6, 0.6]
    ]
    hole1 = Polygon2D.from_array(hole1_verts)

    hole2_verts = [
        [1.3, 0.5],
        [1.5, 0.25],
        [1.1, 0.25]
    ]
    hole2 = Polygon2D.from_array(hole2_verts)

    perim_polys, core_polys = perimeter_core_subpolygons(
        polygon, 0.1, [hole1, hole2], tolerance=1e-5)
    assert len(perim_polys) == 13
    assert len(core_polys) == 2
    assert sum(poly.area for poly in perim_polys + core_polys) == \
        pytest.approx(polygon.area - hole1.area - hole2.area, rel=1e-3)


def test_perimeter_core_subpolygons_signed_zero():
    """Test a case with a signed zero that was failing."""
    polygon_verts = [
        (0.0, -0.0),
        (-66.365025351950564, -64.111893276245013),
        (30.121656801816368, -64.111893276245013),
        (83.372463222004896, 3.5354591123066541),
        (22.759770196994474, -23.559633773719245)
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    perim_polys, core_polys = perimeter_core_subpolygons(polygon, 5.0, tolerance=1e-5)
    assert len(perim_polys) == 5
    for poly in perim_polys:
        assert len(poly) == 4
    assert len(core_polys) == 1
    assert len(core_polys[0]) == 5
    assert sum(poly.area for poly in perim_polys + core_polys) == \
        pytest.approx(polygon.area, rel=1e-3)


def test_perimeter_core_concave_intersect():
    """Test a case where an offset depth creates 4 intersections with a concave shape."""
    polygon_verts = [
        (342.55419132538884, -10.017859160602118),
        (315.77976924081884, 15.204009322638644),
        (304.52663639962361, 49.624331671429445),
        (323.99843057415865, 59.584601981416057),
        (330.48830165015289, 35.040359098751637),
        (345.50412057008964, 17.152954528454099),
        (377.18511138328233, 73.378466870092268),
        (396.15898612775987, 53.204057854261976),
        (351.58292353598546, -18.52304727135888)
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    perim_polys, core_polys = perimeter_core_subpolygons(polygon, 11.0, tolerance=1e-5)
    assert len(perim_polys) == 9
    assert len(core_polys) == 2
    assert sum(poly.area for poly in perim_polys + core_polys) == \
        pytest.approx(polygon.area, rel=1e-3)


def test_perimeter_core_self_intersecting_core():
    """Test a case where the raw extracted core polygon is self-intersecting."""
    polygon_verts = [
        (-37.735447736091537, -14.307192311918065),
        (-37.735447736091537, -7.0708061795016199),
        (-33.572580444862723, -1.9353063435931741),
        (-22.017705814068719, -1.3906321185725812),
        (-16.143005244203756, -9.2884083813711751),
        (-29.837671473292943, -7.2653326884375442),
        (-25.752614785638496, -14.65734002800273),
        (-24.624361033810125, -19.325976242464954),
        (-33.494769841288345, -18.470059603146879)
    ]
    polygon = Polygon2D.from_array(polygon_verts)
    hole_verts = [
        (-35.012076610988572, -13.37346506902562),
        (-29.254091946485165, -14.696245329789917),
        (-33.261338030565241, -9.7552720028173976)
    ]
    hole = Polygon2D.from_array(hole_verts)

    perim_polys, core_polys = perimeter_core_subpolygons(
        polygon, 1.4, [hole], tolerance=1e-5)
    assert len(perim_polys) == 12
    assert len(core_polys) == 2
    assert sum(poly.area for poly in perim_polys + core_polys) == \
        pytest.approx(polygon.area - hole.area, rel=1e-3)
    for core_poly in core_polys:
        assert not core_poly.is_self_intersecting


def test_perimeter_core_subpolygons_infinite_loop_in_core():
    """Test perimeter_core_subpolygons with a core polygraph having a 3-way intersection.

    This can create an infinite loop if the method searching for the core cycle
    has no way to mark already-explored nodes.
    """
    polygon_verts = [
        (192.26116569601979, -80.969790470169656),
        (206.26116569601979, -76.969790470169656),
        (232.26116569601979, -80.969790470169656),
        (232.26116569601979, -40.969790470169656),
        (212.26116569601979, -60.969790470169656),
        (192.26116569601979, -40.969790470169656)
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    hole1_verts = [
        (204.26116569601979, -68.969790470169656),
        (204.26116569601979, -56.969790470169649),
        (212.26116569601979, -64.969790470169656),
        (222.26116569601979, -68.969790470169656)
    ]
    hole1 = Polygon2D.from_array(hole1_verts)

    hole2_verts = [
        (214.26116569601979, -75.969790470169656),
        (218.26116569601979, -70.969790470169656),
        (222.26116569601979, -75.969790470169656)
    ]
    hole2 = Polygon2D.from_array(hole2_verts)

    perim_polys, core_polys = perimeter_core_subpolygons(
        polygon, 1.0, [hole1, hole2], tolerance=1e-5)
    assert len(perim_polys) == 13
    assert len(core_polys) == 2
    assert core_polys[0].polygon_relationship(core_polys[1], 0.01) == 1


def test_perimeter_core_subfaces():
    """Test the perimeter_core_subfaces function."""
    polygon_verts = [[
        [0.,  0., 0.],
        [10., 0., 0.],
        [10., 10., 0.],
        [0., 10., 0.]
    ]]
    face = Face3D.from_array(polygon_verts)

    perim_faces, core_faces = perimeter_core_subfaces(face, 2.0, tolerance=1e-5)
    assert len(perim_faces) == 4
    for f in perim_faces:
        assert len(face) == 4
        assert f.area == pytest.approx(16., rel=1e-3)
    assert len(core_faces) == 1
    assert len(core_faces[0]) == 4
    assert core_faces[0].area == pytest.approx(36., rel=1e-3)
    assert sum(f.area for f in perim_faces + core_faces) == \
        pytest.approx(face.area, rel=1e-3)


def test_perimeter_core_subfaces_and_skeleton():
    """Test the perimeter_core_subfaces_and_skeleton function"""
    polygon_verts = [[
        [0.,  0., 0.],
        [10., 0., 0.],
        [10., 10., 0.],
        [0., 10., 0.]
    ]]
    face = Face3D.from_array(polygon_verts)

    skeleton, perim_faces, core_faces = perimeter_core_subfaces_and_skeleton(
        face, 2.0, tolerance=1e-5)
    assert len(perim_faces) == 4
    for f in perim_faces:
        assert len(face) == 4
        assert f.area == pytest.approx(16., rel=1e-3)
    assert len(core_faces) == 1
    assert len(core_faces[0]) == 4
    assert core_faces[0].area == pytest.approx(36., rel=1e-3)
    assert sum(f.area for f in perim_faces + core_faces) == \
        pytest.approx(face.area, rel=1e-3)
    assert len(skeleton) == 8
    for seg in skeleton:
        assert isinstance(seg, LineSegment3D)
        assert seg.length == pytest.approx(10.0, rel=1e-3) or \
            seg.length == pytest.approx(7.0710678, rel=1e-3)



"""ladybug-geometry-polyskel commands."""
import click
import sys
import logging
import json

from ladybug_geometry.geometry3d import Face3D
from ladybug_geometry_polyskel.polysplit import perimeter_core_subfaces as \
    lgp_perimeter_core_subfaces

_logger = logging.getLogger(__name__)


# command group for all geometry extension commands.
@click.group(help='ladybug geometry polyskel commands.')
@click.version_option()
def main():
    pass


@main.command('perimeter-core-subfaces')
@click.argument('face3d-file', type=click.Path(
    exists=True, file_okay=True, dir_okay=False, resolve_path=True))
@click.argument('perimeter-offset', type=float)
@click.option(
    '--tolerance', '-t', help='The maximum difference between x, y, and z '
    'values at which vertices are considered equivalent.',
    type=float, default=1e-5)
@click.option(
    '--output-file', help='Optional file to output the he string of the visualization '
    'file contents. By default, it will be printed out to stdout',
    type=click.File('w'), default='-', show_default=True)
def perimeter_core_subfaces_cli(face3d_file, perimeter_offset, tolerance, output_file):
    """Compute perimeter and core Face3Ds using the straight skeleton of input Face3Ds.

    \b
    Args:
        face3d_file: Full path to a JSON file containing an array of Face3D
            objects for which core/perimeter Face3Ds will be computed.
        perimeter_offset: Distance to offset perimeter sub-faces.
    """
    try:
        perimeter_core_subfaces(face3d_file, perimeter_offset, tolerance, output_file)
    except Exception as e:
        _logger.exception('Failed to compute core/perimeter Face3Ds.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


def perimeter_core_subfaces(
        face3d_file, perimeter_offset, tolerance=1e-5, output_file=None
):
    """Compute perimeter and core Face3Ds using the straight skeleton of input Face3Ds.

    Args:
        face3d_file: Full path to a JSON file containing an array of Face3D
            objects for which core/perimeter Face3Ds will be computed.
        perimeter_offset: Distance to offset perimeter sub-faces.
        tolerance: The maximum difference between x, y, and z values at which
            vertices are considered equivalent.
        output_file: Optional file to output the string of the visualization
            file contents. If None, the string will simply be returned from
            this method.
    """
    # load the Face3D objects
    with open(face3d_file) as inf:
        data = json.load(inf)
    face_3ds = [Face3D.from_dict(geo) for geo in data]

    # compute core/perimeter Face3Ds
    core_per_geos = []
    for geo_obj in face_3ds:
        perimeter_sub_faces, core_sub_faces = \
            lgp_perimeter_core_subfaces(geo_obj, perimeter_offset, tolerance)
        core_per_geos.extend([f.to_dict() for f in perimeter_sub_faces])
        core_per_geos.extend([f.to_dict() for f in core_sub_faces])

    # write the resulting geometries into the output
    if output_file is None:
        return json.dumps(core_per_geos)
    else:
        output_file.write(json.dumps(core_per_geos))



"""Test cli."""
from click.testing import CliRunner
import json

from ladybug_geometry.geometry3d import Face3D
from ladybug_geometry_polyskel.cli import perimeter_core_subfaces_cli


def test_perimeter_core_subfaces():
    input_file = './tests/assets/sample_geometries.json'
    runner = CliRunner()
    result = runner.invoke(
        perimeter_core_subfaces_cli, [input_file, '1.0', '--tolerance', '0.01'])
    assert result.exit_code == 0

    result_dict = json.loads(result.output)
    new_faces = [Face3D.from_dict(f) for f in result_dict]
    assert len(new_faces) == 33



# coding=utf-8
"""Test the methods for splitting polygons with a straight skeleton."""
from __future__ import division
import pytest

from ladybug_geometry.geometry2d import Polygon2D

from ladybug_geometry_polyskel.offset import offset_polygon, offset_perimeter_polygons


def test_offset_square():
    """Test the offset_polygon function with a square."""
    polygon_verts = [
        [0.,  0.],
        [10., 0.],
        [10., 10.],
        [0., 10.]
    ]
    polygon = Polygon2D.from_array(polygon_verts)

    core_polys = offset_polygon(polygon, 2.0, tolerance=1e-5)
    assert len(core_polys) == 1
    assert len(core_polys[0]) == 4
    assert core_polys[0].area == pytest.approx(36., rel=1e-3)

    core_polys = offset_polygon(polygon, -2.0, tolerance=1e-5)
    assert len(core_polys) == 1
    assert len(core_polys[0]) == 4
    assert core_polys[0].area == pytest.approx(196., rel=1e-3)

    perim_polys = offset_perimeter_polygons(polygon, 2.0, tolerance=1e-5)
    assert len(perim_polys) == 4
    for poly in perim_polys:
        assert len(poly) == 4
        assert poly.area == pytest.approx(16., rel=1e-3)

    perim_polys = offset_perimeter_polygons(polygon, -2.0, tolerance=1e-5)
    assert len(perim_polys) == 4
    for poly in perim_polys:
        assert len(poly) == 4
        assert poly.area == pytest.approx(24., rel=1e-3)



# coding=utf-8
"""Functions for splitting a polygon into sub-polygons based on skeleton topology."""
from __future__ import division

from ladybug_geometry.geometry2d import LineSegment2D, Polygon2D
from ladybug_geometry.geometry3d import Vector3D, LineSegment3D, Face3D

from .polygraph import PolygonDirectedGraph, skeleton_as_directed_graph


def perimeter_core_subfaces(face, distance, tolerance=1e-5):
    """Compute perimeter and core Face3Ds using the straight skeleton of an input Face3D.

    Args:
        face: A Face3D to split into perimeter and core sub-faces.
        distance: Distance to offset perimeter sub-faces.
        tolerance: Tolerance for point equivalence. (Default: 1e-5).

    Returns:
        A tuple with two items.

        * perimeter_sub_faces -- A list of Face3Ds for perimeter sub-faces.

        * core_sub_faces -- A list of Face3D for core sub-faces.
    """
    # get core and perimeter sub-polygons
    perimeter, core = perimeter_core_subpolygons(
        face.boundary_polygon2d, distance, face.hole_polygon2d, tolerance, False)
    # convert the polygons into Face3Ds
    return _polygons_to_face3d(face, perimeter, core)


def perimeter_core_subfaces_and_skeleton(face, distance, tolerance=1e-5):
    """Compute perimeter and core Face3Ds using the straight skeleton of an input Face3D.

    Args:
        face: A Face3D to split into perimeter and core sub-faces.
        distance: Distance to offset perimeter sub-faces.
        tolerance: Tolerance for point equivalence. (Default: 1e-5).

    Returns:
        A tuple with two items.

        * skeleton -- A list of LineSegment3D for the segments of the straight
            skeleton being used to generate core/perimeter polygons.

        * perimeter_sub_faces -- A list of Face3Ds for perimeter sub-faces.

        * core_sub_faces -- A list of Face3D for core sub-faces.
    """
    # get core and perimeter sub-polygons
    skeleton_2d, perimeter, core = perimeter_core_subpolygons_and_skeleton(
        face.boundary_polygon2d, distance, face.hole_polygon2d, tolerance)
    perimeter_sub_faces, core_sub_faces = _polygons_to_face3d(face, perimeter, core)
    # convert the skeleton segments into LineSegment3Ds
    skeleton = []
    for seg in skeleton_2d:
        verts_3d = tuple(face.plane.xy_to_xyz(pt) for pt in seg.vertices)
        skeleton.append(LineSegment3D.from_end_points(*verts_3d))
    return skeleton, perimeter_sub_faces, core_sub_faces


def perimeter_core_subpolygons(
        polygon, distance, holes=None, tolerance=1e-5, flat_core=True):
    """Compute perimeter and core sub-polygons using the polygon straight skeleton.

    Args:
        polygon: A Polygon2D to split into perimeter and core sub-polygons.
        distance: Distance to offset perimeter sub-polygons.
        holes: A list of Polygon2D objects representing holes in the
            polygon. (Default: None).
        tolerance: Tolerance for point equivalence. (Default: 1e-5).
        flat_core: A boolean to note whether the core_sub_polys should be returned
            as a flat list or as a nested list of lists for boundaries and
            holes within each geometry.

    Returns:
        A tuple with two items.

        * perimeter_sub_polys -- A list of Polygon2Ds for perimeter sub-polygons.

        * core_sub_polys -- When flat_core is True, this will be a list of Polygon2D
            where each Polygon2D represents a loop in the core geometry. When
            flat_core is False, this will be a list of lists where each sub-list
            contains Polygon2Ds and represents one core geometry. The sub-list
            will have has at least one Polygon2D in it and, in the event that
            a core geometry has holes, there will be multiple Polygon2Ds in the
            sub-list. The first item in the list will be the outer boundary of
            the geometry and successive items represent hole polygons.
    """
    # initialize sub polygon lists
    perimeter_sub_polys, core_sub_polys = [], []
    # compute the straight skeleton of the polygon as a directed graph
    dg = skeleton_as_directed_graph(polygon, holes, tolerance)
    # traverse the directed graph to get all of the perimeter polygons
    _perimeter_sub_dg = None
    root_keys = [dg.outer_root_key] + dg.hole_root_keys
    for root_key in root_keys:
        # compute the polygons on the perimeter of the polygon
        _perimeter_sub_polys, _perimeter_sub_dg = _split_perimeter_subpolygons(
            dg, distance, root_key, tolerance, _perimeter_sub_dg)
        perimeter_sub_polys.extend(_perimeter_sub_polys)  # collect perimeter sub-polys
    # compute the polygons on the core of the polygon
    core_sub_polys = _exterior_cycles_as_polygons(_perimeter_sub_dg, tolerance)
    if not flat_core and len(core_sub_polys) != 0:  # remake cores w/ holes
        core_sub_polys = Polygon2D.group_boundaries_and_holes(core_sub_polys, tolerance)
    return perimeter_sub_polys, core_sub_polys


def perimeter_core_subpolygons_and_skeleton(
        polygon, distance, holes=None, tolerance=1e-5):
    """Get perimeter and core sub-polygons with line segments for the straight skeleton.

    Args:
        polygon: A Polygon2D to split into perimeter and core sub-polygons.
        distance: Distance to offset perimeter sub-polygons.
        holes: A list of Polygon2D objects representing holes in the
            polygon. (Default: None).
        tolerance: Tolerance for point equivalence. (Default: 1e-5).

    Returns:
        A tuple with three items.

        * skeleton -- A list of LineSegment2D for the segments of the straight
            skeleton being used to generate core/perimeter polygons.

        * perimeter_sub_polys -- A list of Polygon2Ds for perimeter sub-polygons.

        * core_sub_polys -- A list of lists where each sub-list contains Polygon2Ds
            and represents one core geometry. The sub-list will have has at
            least one Polygon2D and, in the event that a core geometry has holes,
            there will be multiple Polygon2Ds in the sub-list. The first item in
            the list will be the outer boundary of the geometry and successive
            items represent hole polygons.
    """
    # initialize sub polygon lists
    perimeter_sub_polys, core_sub_polys = [], []
    # compute the straight skeleton of the polygon as a directed graph
    dg = skeleton_as_directed_graph(polygon, holes, tolerance)
    skeleton = dg.connection_segments
    # traverse the directed graph to get all of the perimeter polygons
    _perimeter_sub_dg = None
    root_keys = [dg.outer_root_key] + dg.hole_root_keys
    for root_key in root_keys:
        # compute the polygons on the perimeter of the polygon
        _perimeter_sub_polys, _perimeter_sub_dg = _split_perimeter_subpolygons(
            dg, distance, root_key, tolerance, _perimeter_sub_dg)
        perimeter_sub_polys.extend(_perimeter_sub_polys)  # collect perimeter sub-polys
    # compute the polygons on the core of the polygon
    core_sub_polys = _exterior_cycles_as_polygons(_perimeter_sub_dg, tolerance)
    if len(core_sub_polys) != 0:  # remake cores w/ holes
        core_sub_polys = Polygon2D.group_boundaries_and_holes(core_sub_polys, tolerance)
    return skeleton, perimeter_sub_polys, core_sub_polys


def _split_perimeter_subpolygons(dg, distance, root_key, tol, core_graph=None):
    """Split polygons in the PolygonDirectedGraph into perimeter sub-polygons.

    Args:
        dg: A PolygonDirectedGraph object that comes from the skeleton_as_directed_graph
            function and contains separations for how a geometry will be split.
        distance: The distance to offset the perimeter in model units.
        root_key: A string key for a node in the input dg that identifies the
            perimeter loop to be traversed. This will be used as the starting
            point for graph traversal.
        tol: A number representing the tolerance for point equivalence.
        core_graph: An optional PolygonDirectedGraph that is being built to extract
            the core of the shape after all perimeter sub-polygons are extracted.
            This is used when the input dg contains holes and this method for
            _split_perimeter_subpolygons must be run multiple times for the
            boundary and each hole. (Default: None).

    Returns:
        A tuple with two lists:

        * perimeter_subpolygons: A list of perimeter sub-polygons as Polygon2D objects.

        * core_graph: A PolygonDirectedGraph with just the perimeter sub-polygons.
            The edges of this graph that are not bidirectional (exterior edges)
            define the exterior of the polygon and it's core sub-polygons.
    """
    # set up a list lto collect the perimeter polygons and the core graph
    perimeter_subpolygons = []
    core_graph = PolygonDirectedGraph(tol=tol) if core_graph is None else core_graph
    exter_cycle = dg.exterior_cycle(dg.node(root_key))

    # cycle through exterior nodes and get split polygons using the distance
    for i, base_node in enumerate(exter_cycle):
        try:
            next_node = exter_cycle[i + 1]
        except IndexError:  # the last edge of the polygon
            next_node = exter_cycle[0]
        # find the smallest polygon defined by the exterior node
        min_ccw_poly_graph = dg.min_cycle(next_node, base_node)

        # proceed to extract the polygon and build up the core graph
        if min_ccw_poly_graph is not None:
            # everything is good, offset edge by distance and cut a perimeter polygon
            split_poly_graph = _split_polygon_graph(
                base_node, next_node, distance, min_ccw_poly_graph, tol)
            # store the perimeter nodes in the core_graph
            pts = [node.pt for node in split_poly_graph]
            for i in range(len(pts) - 1):
                core_graph.add_node(pts[i], [pts[i + 1]])
            # store perimeter points in list as polygon
            perimeter_subpolygons.append(Polygon2D(pts))

        else:  # handle the case of unsuccessful minimum cycle polygon extraction
            # if we are just missing a skeleton segment, there may still be hope
            if len(next_node.adj_lst) == 1:  # connect the node to the nearest interior
                rel_nodes, node_dists = [], []
                for node in dg.nodes:
                    if not node.exterior:
                        rel_nodes.append(node)
                        node_dists.append(next_node.pt.distance_to_point(node.pt))
                near_nodes = [n for _, n in sorted(zip(node_dists, rel_nodes),
                                                   key=lambda pair: pair[0])]
                dg.add_adj(next_node, [near_nodes[0].pt])
                dg.add_adj(near_nodes[0], [next_node.pt])  # add node di-directionally
                min_ccw_poly_graph = dg.min_cycle(next_node, base_node)

            if min_ccw_poly_graph is not None:  # connecting the node worked!
                split_poly_graph = _split_polygon_graph(
                    base_node, next_node, distance, min_ccw_poly_graph, tol)
                pts = [node.pt for node in split_poly_graph]
                for i in range(len(pts) - 1):
                    core_graph.add_node(pts[i], [pts[i + 1]])
                perimeter_subpolygons.append(Polygon2D(pts))

            else:  # no hope of extracting a perimeter sub-polygon here
                msg = 'Failed to compute sub-polygons between ' \
                    'Node {} and Node {}.'.format(base_node, next_node)
                print(msg)
                # close the core graph with the exterior segment and move on
                core_graph.add_node(next_node.pt, [base_node.pt])

    return perimeter_subpolygons, core_graph


def _split_polygon_graph(node1, node2, distance, poly_graph, tol):
    """Split the PolygonDirectedGraph by an edge offset at a distance.

    Args:
        node1: A node defining the edge to offset for splitting. If the graph
            defines a counter-clockwise polygon, this node should represent
            the left hand-most point.
        node2: A node defining the edge to offset for splitting. If the graph
            defines a counter-clockwise polygon, this node should represent
            the right hand-most point.
        distance: The distance to offset the splitting edge, in model units.
        poly_graph: A list of Nodes representing a PolygonDirectedGraph for
            a single polygon.
        tol: Tolerance for point equivalence.

    Returns:
        A list of nodes representing the split polygon adjacent to the exterior
        edge defined by node1 and node2.
    """
    # use the nodes to create a line segment
    ext_seg = LineSegment2D.from_end_points(node1.pt, node2.pt)
    # compute normal facing into the polygon
    ext_arr = ext_seg.v.to_array()
    ext_seg_v = Vector3D(ext_arr[0], ext_arr[1], 0)
    normal = ext_seg_v.cross(Vector3D(0, 0, -1)).normalize() * distance

    # move the segment by normal to get offset segment
    offset_seg = ext_seg.move(normal)
    poly_graph = PolygonDirectedGraph.from_point_array(
        [n.pt for n in poly_graph], tol=tol)
    ordered_nodes = poly_graph.ordered_nodes
    poly_graph.outer_root_key = ordered_nodes[0].key
    root_node = ordered_nodes[0]
    goal_node = ordered_nodes[-1]

    # update graph by intersecting offset segment with other edges
    poly_graph.intersect_graph_with_segment(offset_seg)

    # get the minimum cycle through the graph connecting the nodes
    # since we start at the exterior edge, this will include the perimeter offset
    split_poly_nodes = poly_graph.min_cycle(root_node, goal_node, ccw_only=True)

    return split_poly_nodes


def _exterior_cycles_as_polygons(dg, tol):
    """Convert and sort exterior cycles in a PolygonDirectedGraph to a list of polygons.

    Args:
        dg: A PolygonDirectedGraph.
        tol: The tolerance at which point equivalence is set.

    Returns:
        A list of Polygon2D objects sorted by area.
    """
    ext_cycles = dg.exterior_cycles()
    ext_polygons = []
    for cycle in ext_cycles:
        ext_poly = Polygon2D([node.pt for node in cycle])
        ext_poly = ext_poly.remove_colinear_vertices(tol)
        if ext_poly.is_self_intersecting:
            ext_polygons.extend(ext_poly.split_through_self_intersection(tol))
        else:
            ext_polygons.append(ext_poly)
    return ext_polygons


def _polygons_to_face3d(face, perimeter, core):
    """Convert lists of perimeter and core Polygon2D into Face3D."""
    # convert the perimeter polygons into Face3Ds
    perimeter_sub_faces = []
    for poly in perimeter:
        verts_3d = tuple(face.plane.xy_to_xyz(pt) for pt in poly.vertices)
        perimeter_sub_faces.append(Face3D(verts_3d, face.plane))
    # convert the core polygons into Face3Ds
    core_sub_faces = []
    for poly_group in core:
        all_verts_3d = [tuple(face.plane.xy_to_xyz(pt) for pt in poly.vertices)
                        for poly in poly_group]
        if len(all_verts_3d) == 1:  # no holes in the shape
            core_sub_faces.append(Face3D(all_verts_3d[0], face.plane))
        else:
            core_sub_faces.append(Face3D(all_verts_3d[0], face.plane, all_verts_3d[1:]))
    return perimeter_sub_faces, core_sub_faces



# coding=utf-8
"""Functions for offsetting a polygon using the straight skeleton."""
from __future__ import division
import math

from ladybug_geometry.geometry2d import Point2D, Ray2D, Polygon2D

from .polygraph import skeleton_as_directed_graph
from .polysplit import perimeter_core_subpolygons, _split_perimeter_subpolygons, \
    _exterior_cycles_as_polygons


def offset_polygon(polygon, distance, tolerance=1e-5):
    """Offset a Polygon2D by a given distance inwards or outwards.

    Assuming that the input is not self-intersecting, the resulting shapes
    should have all self-intersections resolved through the straight skeleton.
    However, if the distance is large enough, the result may be an empty list.

    Args:
        polygon: A Polygon2D to be offset inwards or outwards by a given distance.
        distance: The distance inwards that the polygon will be offset.
            Positive values will always be offset inwards while negative ones
            will be offset outwards.
        tolerance: Tolerance for point equivalence. (Default: 1e-5).
    """
    # determine the methods to use based on the offset direction
    if -tolerance <= distance <= tolerance:
        return [polygon]
    elif distance < 0:  # we are offsetting outwards
        offset_poly = polygon.offset(distance, check_intersection=False)
        if not offset_poly.is_self_intersecting:
            return [offset_poly]
        bound_rect = _bound_rect_for_ext_skeleton(polygon, distance)
        dg = skeleton_as_directed_graph(bound_rect, [polygon], tolerance)
        root_key = dg.hole_root_keys[0]
        _, _perimeter_sub_dg = _split_perimeter_subpolygons(
            dg, abs(distance), root_key, tolerance, None)
        offset_polygons = _exterior_cycles_as_polygons(_perimeter_sub_dg, tolerance)
    else:  # we are offsetting inwards
        _, offset_polygons = perimeter_core_subpolygons(
            polygon, distance, tolerance=tolerance, flat_core=True)
    return offset_polygons


def offset_perimeter_polygons(polygon, distance, tolerance=1e-5):
    """Get a list of Polygon2Ds along the perimeter of an input polygon.

    These can be used to split the input polygons into core/perimeter shapes.
    Assuming that the input is not self-intersecting, the resulting shapes
    should have all self-intersections resolved through the straight skeleton.

    Args:
        polygon: A Polygon2D to be offset inwards or outwards by a given distance.
        distance: The distance inwards that the polygon will be offset.
            Positive values will always be offset inwards while negative ones
            will be offset outwards.
        tolerance: Tolerance for point equivalence. (Default: 1e-5).
    """
    # determine the methods to use based on the offset direction
    if -tolerance <= distance <= tolerance:
        return []
    elif distance < 0:  # we are offsetting outwards
        offset_poly = polygon.offset(distance, check_intersection=False)
        if not offset_poly.is_self_intersecting:
            perimeter_polygons = []
            for i, (in_pt, out_pt) in enumerate(zip(polygon, offset_poly)):
                pts = (polygon[i - 1], offset_poly[i - 1], out_pt, in_pt)
                perimeter_polygons.append(Polygon2D(pts))
            return perimeter_polygons
        bound_rect = _bound_rect_for_ext_skeleton(polygon, distance)
        dg = skeleton_as_directed_graph(bound_rect, [polygon], tolerance)
        root_key = dg.hole_root_keys[0]
        perimeter_polygons, _ = _split_perimeter_subpolygons(
            dg, abs(distance), root_key, tolerance, None)
    else:  # we are offsetting inwards
        dg = skeleton_as_directed_graph(polygon, None, tolerance)
        root_key = dg.outer_root_key
        perimeter_polygons, _ = _split_perimeter_subpolygons(
            dg, distance, root_key, tolerance, None)
    return perimeter_polygons


def offset_skeleton(polygon, distance, tolerance=1e-5):
    """Get a list of LineSegment2D for the straight used to offset a polygon.

    Args:
        polygon: A Polygon2D to be offset inwards or outwards by a given distance.
        distance: The distance inwards that the polygon will be offset.
            Positive values will always be offset inwards while negative ones
            will be offset outwards.
        tolerance: Tolerance for point equivalence. (Default: 1e-5).
    """
    if -tolerance <= distance <= tolerance:
        return []
    elif distance < 0:  # we are offsetting outwards
        bound_rect = _bound_rect_for_ext_skeleton(polygon, distance)
        dg = skeleton_as_directed_graph(bound_rect, [polygon], tolerance)
    else:  # we are offsetting inwards
        dg = skeleton_as_directed_graph(polygon, None, tolerance)
    return dg.connection_segments


def _bound_rect_for_ext_skeleton(polygon, distance):
    """Get a bounding rectangle to be used for offsetting a polygon outwards.

    Args:
        polygon: A Polygon2D that is being offset outwards for which a bounding
            rectangle is needed.
        distance: The distance outwards that the polygon will be offset. This
            value should always be negative.

    Returns:
        A Polygon2D for the bounding rectangle around the shape.
    """
    # determine how far from the polygon the bounding rectangle will be drawn
    min_pt, max_pt = polygon.min, polygon.max
    poly_dim = min((max_pt.x - min_pt.x, max_pt.y - min_pt.y))
    offset_factor = 50 if abs(distance) < poly_dim / 10 else 10
    offset = abs(distance) * offset_factor

    # create a starting bounding rectangle from the min, max, and distance
    min_pt = Point2D(min_pt.x - offset, min_pt.y - offset)
    max_pt = Point2D(max_pt.x + offset, max_pt.y + offset)
    min_x_max_y, max_x_min_y = Point2D(min_pt.x, max_pt.y), Point2D(max_pt.x, min_pt.y)
    rect_pts = (min_pt, max_x_min_y, max_pt, min_x_max_y)
    rect_poly = Polygon2D(rect_pts)

    # find a corner point to be used as the base for the rectangle
    pts_dists, pts_info = [], []
    for corner_i, conrner_pt in enumerate(rect_poly):
        for pt_i, pt in enumerate(polygon.vertices):
            pts_dists.append(conrner_pt.distance_to_point(pt))
            pts_info.append((pt_i, corner_i))
    sort_info = [inf for _, inf in sorted(zip(pts_dists, pts_info),
                                          key=lambda pair: pair[0])]
    poly_pt_i, corner_pt_i = sort_info[0]

    # shoot a ray from the polygon point to the rectangle
    poly_pt = polygon.vertices[poly_pt_i]
    v1 = polygon.vertices[poly_pt_i - 1] - poly_pt
    end_i = poly_pt_i + 1 if poly_pt_i != len(polygon) - 1 else 0
    v2 = polygon.vertices[end_i] - poly_pt
    if not polygon.is_clockwise:
        ang = v1.angle_clockwise(v2) / 2
        if ang == 0:
            ang = math.pi / 2
        m_vec = v1.rotate(-ang).normalize()
        m_dist = -distance / math.sin(ang)
    else:
        ang = v1.angle_counterclockwise(v2) / 2
        if ang == 0:
            ang = math.pi / 2
        m_vec = v1.rotate(ang).normalize()
        m_dist = distance / math.sin(ang)
    m_vec = m_vec * m_dist
    corner_ray = Ray2D(poly_pt, m_vec)
    ray_int = rect_poly.intersect_line_ray(corner_ray)

    # if the ray intersection is too far from the corner, just use the start rectangle
    corner_pt = rect_pts[corner_pt_i]
    safe_distance = abs(distance) * (offset_factor - 3)
    if len(ray_int) == 0 or ray_int[0].distance_to_point(corner_pt) > safe_distance:
        return rect_poly

    # otherwise, adjust the bounding rectangle to give us a nicer skeleton
    bpt = ray_int[0]
    base, height = max_pt.x - min_pt.x, max_pt.y - min_pt.y
    if corner_pt_i == 0:  # lower left corner
        rect_pts = (
            bpt, Point2D(bpt.x + base, bpt.y), Point2D(bpt.x + base, bpt.y + height),
            Point2D(bpt.x, bpt.y + height))
    elif corner_pt_i == 1:  # lower right corner
        rect_pts = (
            Point2D(bpt.x - base, bpt.y), bpt, Point2D(bpt.x, bpt.y + height),
            Point2D(bpt.x - base, bpt.y + height))
    elif corner_pt_i == 2:  # upper right corner
        rect_pts = (
            Point2D(bpt.x - base, bpt.y - height), Point2D(bpt.x, bpt.y - height),
            bpt, Point2D(bpt.x - base, bpt.y))
    else:  # upper left corner
        rect_pts = (
            Point2D(bpt.x, bpt.y - height), Point2D(bpt.x + base, bpt.y - height),
            Point2D(bpt.x + base, bpt.y), bpt)
    return Polygon2D(rect_pts)
